import { renderItem } from './render-utils.mjs';
import { generatorToString } from './utils.js';

/** @import { HtmlMethodResult } from './element-types.js' */

/**
 * @yields {string}
 * @param {HtmlMethodResult} item
 * @returns {AsyncIterableIterator<string>}
 */
export const render = async function * (item) {
  if (item === undefined) throw new TypeError('Expected an argument');
  if (!item) throw new TypeError(`Expected a non-falsy argument, got: ${item}`);
  if (Array.isArray(item)) {
    for (const value of item) {
      yield * render(value);
    }
  } else if (typeof item === 'string' || typeof item === 'object') {
    yield * renderItem(item);
  } else {
    throw new TypeError(`Expected a string or an object, got: ${typeof item}`);
  }
};

/**
 * @param {HtmlMethodResult} item
 * @returns {Promise<string>}
 */
export const renderToString = async (item) => generatorToString(render(item));
